from typing import Generic, TypeVar

from rainbowadn.chain.block import Block, BlockFactory
from rainbowadn.chain.blockchainprotocol import BlockChainProtocol
from rainbowadn.chain.chaincollectionfactory import ChainCollectionFactory
from rainbowadn.chain.chaincollectioninterface import ChainCollectionInterface
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference
from rainbowadn.hashing.rainbow_factory import RainbowFactory

__all__ = ('BlockChain', 'BlockChainFactory',)

HeaderType = TypeVar('HeaderType')
StateType = TypeVar('StateType')
ActualStateType = TypeVar('ActualStateType')


class BlockChain(
    ChainCollectionInterface[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ],
    Generic[HeaderType, ActualStateType, StateType],
):
    def __init__(
            self,
            reference: NullableReference[
                Block[
                    HeaderType,
                    StateType
                ]
            ],
            protocol: BlockChainProtocol[HeaderType, StateType, ActualStateType],
    ):
        assert isinstance(reference, NullableReference)
        assert isinstance(protocol, BlockChainProtocol)
        super().__init__(reference)
        self.protocol = protocol

    def factory(self) -> ChainCollectionFactory[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ]:
        return BlockChainFactory(self.protocol)

    def add(
            self,
            header: HashPoint[HeaderType]
    ) -> ChainCollectionInterface[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ]:
        assert isinstance(header, HashPoint)
        return self.factory().from_reference(
            NullableReference.off(
                self._add(header)
            )
        )

    def state(
            self
    ) -> NullableReference[
        StateType
    ]:
        if isinstance(self.reference.reference, Null):
            return NullableReference(
                Null(),
                self.protocol.state_factory
            )
        elif isinstance(self.reference.reference, NotNull):
            block: Block[
                HeaderType,
                StateType
            ] = self.reference.reference.value.resolve()
            assert isinstance(block, Block)
            return NullableReference.of(block.state)
        else:
            raise TypeError

    def _add(
            self,
            header: HashPoint[HeaderType]
    ) -> Block[
        HeaderType,
        StateType
    ]:
        assert isinstance(header, HashPoint)
        return Block(
            self.reference,
            header,
            self.protocol.state_protocol.derive(
                self.state(),
                header,
            )
        )

    def previous(
            self
    ) -> ChainCollectionInterface[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ]:
        assert isinstance(self.reference.reference, NotNull)
        block: Block[
            HeaderType,
            StateType
        ] = self.reference.reference.value.resolve()
        assert isinstance(block, Block)
        return self.factory().from_reference(block.previous)

    def verify(self) -> bool:
        if isinstance(self.reference.reference, Null):
            return True
        elif isinstance(self.reference.reference, NotNull):
            return self.verify_link(
                self.previous().reference
            ) and self.previous().verify()
        else:
            raise TypeError

    def _verify_link(
            self,
            block: Block[
                HeaderType,
                StateType
            ],
            previous: NullableReference[
                Block[
                    HeaderType,
                    StateType
                ]
            ]
    ) -> bool:
        assert isinstance(block, Block)
        assert isinstance(previous, NullableReference)
        assert block.previous == previous
        previous_chain = self.factory().from_reference(previous)
        assert isinstance(previous_chain, BlockChain)
        return self.protocol.state_protocol.verify(
            previous_chain.state(),
            block.header,
            block.state,
        )

    def _actual_state_factory(self) -> RainbowFactory[ActualStateType]:
        return self.protocol.actual_state_factory()

    def _actual_state(
            self,
            block: Block[
                HeaderType,
                StateType
            ]
    ) -> HashPoint[ActualStateType]:
        assert isinstance(block, Block)
        return self.protocol.actual_state(block.state.resolve())

    def loose(self) -> ChainCollectionInterface[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ]:
        return self


class BlockChainFactory(
    ChainCollectionFactory[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ],
    Generic[HeaderType, ActualStateType, StateType]
):
    def __init__(
            self,
            protocol: BlockChainProtocol[HeaderType, StateType, ActualStateType],
    ):
        assert isinstance(protocol, BlockChainProtocol)
        self.protocol = protocol

    def empty(self) -> BlockChain[HeaderType, ActualStateType, StateType]:
        return BlockChain(
            NullableReference(
                Null(),
                BlockFactory(
                    self.protocol.header_factory,
                    self.protocol.state_factory
                )
            ),
            self.protocol
        )

    def from_reference(
            self,
            reference: NullableReference[
                Block[
                    HeaderType,
                    StateType
                ],
            ]
    ) -> BlockChain[
        HeaderType, ActualStateType, StateType
    ]:
        assert isinstance(reference, NullableReference)
        return BlockChain(
            reference,
            self.protocol
        )
