from typing import Generic, TypeVar

from rainbowadn.chain.stages.derivation.derived import Derived
from rainbowadn.chain.stages.derivation.derivedstage import DerivedStage
from rainbowadn.chain.stages.derivation.derivedstate import DerivedState
from rainbowadn.chain.stages.stage import StageStage, StageStageFactory, StateStage
from rainbowadn.chain.stages.stageprotocol import StageProtocol
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference
from rainbowadn.hashing.rainbow_factory import RainbowFactory

__all__ = ('ActiveStageProtocol',)

HeaderType = TypeVar('HeaderType')
BaseStateType = TypeVar('BaseStateType')
StageType = TypeVar('StageType')


class ActiveStageProtocol(
    StageProtocol[HeaderType, BaseStateType, StageType],
    Generic[HeaderType, BaseStateType, StageType]
):
    def derive_header(
            self,
            previous: NullableReference[BaseStateType],
            header: HashPoint[HeaderType],
    ) -> HashPoint[StageType]:
        raise NotImplementedError

    def derive_stage_or_state(
            self,
            stage: HashPoint[StageType],
    ) -> Derived[BaseStateType, StageType]:
        raise NotImplementedError

    @classmethod
    def _previous_base_state(
            cls,
            previous_reference: NullableReference[StateStage[HeaderType, BaseStateType, StageType]],
            base_state_factory: RainbowFactory[BaseStateType],
    ) -> NullableReference[BaseStateType]:
        assert isinstance(previous_reference, NullableReference)
        assert isinstance(base_state_factory, RainbowFactory)
        if isinstance(previous_reference.reference, Null):
            return NullableReference(Null(), base_state_factory)
        elif isinstance(previous_reference.reference, NotNull):
            previous_state_stage: StateStage[HeaderType, BaseStateType, StageType] = (
                previous_reference.reference.value.resolve()
            )
            assert isinstance(previous_state_stage, StateStage)
            return NullableReference.of(previous_state_stage.state)
        else:
            raise TypeError

    def _derive_cycle(
            self,
            stage: StageStage[HeaderType, BaseStateType, StageType],
            stage_factory: RainbowFactory[StageType]
    ) -> HashPoint[StateStage[HeaderType, BaseStateType, StageType]]:
        assert isinstance(stage, StageStage)
        assert isinstance(stage_factory, RainbowFactory)
        while True:
            derived = self.derive_stage_or_state(stage.stage)
            assert isinstance(derived, Derived)
            if isinstance(derived, DerivedStage):
                stage = StageStage(self, NullableReference.off(stage), derived.stage)
                assert isinstance(stage, StageStage)
            elif isinstance(derived, DerivedState):
                return HashPoint.of(StateStage(self, HashPoint.of(stage), derived.state, stage_factory))
            else:
                raise TypeError

    def _derive_initial_stage(
            self,
            previous: NullableReference[BaseStateType],
            header: HashPoint[HeaderType],
            stage_factory: RainbowFactory[StageType],
    ) -> StageStage[HeaderType, BaseStateType, StageType]:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        assert isinstance(stage_factory, RainbowFactory)
        return StageStage(
            self,
            NullableReference(
                Null(),
                StageStageFactory(
                    self,
                    stage_factory
                )
            ),
            self.derive_header(previous, header)
        )

    def derive_full(
            self,
            previous_reference: NullableReference[StateStage[HeaderType, BaseStateType, StageType]],
            header: HashPoint[HeaderType],
            base_state_factory: RainbowFactory[BaseStateType],
            stage_factory: RainbowFactory[StageType],
    ) -> HashPoint[StateStage[HeaderType, BaseStateType, StageType]]:
        assert isinstance(previous_reference, NullableReference)
        assert isinstance(header, HashPoint)
        assert isinstance(base_state_factory, RainbowFactory)
        assert isinstance(stage_factory, RainbowFactory)
        return self._derive_cycle(
            self._derive_initial_stage(
                self._previous_base_state(
                    previous_reference,
                    base_state_factory
                ),
                header,
                stage_factory
            ),
            stage_factory
        )

    def verify_header(
            self,
            previous: NullableReference[BaseStateType],
            header: HashPoint[HeaderType],
            stage: HashPoint[StageType]
    ) -> bool:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        assert isinstance(stage, HashPoint)
        assert stage == self.derive_header(previous, header)
        return True

    def verify_stage(
            self,
            previous: HashPoint[StageType],
            stage: HashPoint[StageType]
    ) -> bool:
        assert isinstance(previous, HashPoint)
        assert isinstance(stage, HashPoint)
        derived = self.derive_stage_or_state(previous)
        assert isinstance(derived, Derived)
        if isinstance(derived, DerivedStage):
            assert stage == derived.stage
            return True
        else:
            raise TypeError

    def verify_state(
            self,
            stage: HashPoint[StageType],
            state: HashPoint[BaseStateType]
    ) -> bool:
        assert isinstance(stage, HashPoint)
        assert isinstance(state, HashPoint)
        derived = self.derive_stage_or_state(stage)
        assert isinstance(derived, Derived)
        if isinstance(derived, DerivedState):
            assert state == derived.state
            return True
        else:
            raise TypeError
