from typing import Generic, TypeVar

from rainbowadn.data.collection.collection_interface.activecollectioninterface import ActiveCollectionInterface
from rainbowadn.data.collection.collection_interface.querycollectioninterface import QueryCollectionInterface
from rainbowadn.data.collection.keymetadata import KeyMetadata, KeyMetadataFactory
from rainbowadn.data.collection.keymetadataquerycollection import KeyMetadataQueryCollection
from rainbowadn.data.collection.trees.binary.balancedtreecreationprotocol import BalancedTreeCreationProtocol
from rainbowadn.data.collection.trees.binary.binarytree import BinaryTree, BinaryTreeFactory
from rainbowadn.data.collection.trees.binary.binarytreebalancingprotocol import BinaryTreeBalancingProtocol
from rainbowadn.data.collection.trees.binary.binarytreecreationprotocol import BinaryTreeCreationProtocol
from rainbowadn.data.collection.trees.binary.querybinarytree import QueryBinaryTree
from rainbowadn.data.collection.trees.comparison.comparator import (Comparator, Comparison, Equal, Left, Replace, Right)
from rainbowadn.data.collection.trees.comparison.keyedcomparator import KeyedComparator
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference
from rainbowadn.hashing.rainbow_factory import RainbowFactory

__all__ = ('ActiveBinaryTree',)

ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')


class ActiveBinaryTree(
    ActiveCollectionInterface[BinaryTree[KeyMetadata[ActiveKeyType, MetaDataType]], ActiveKeyType],
    Generic[ActiveKeyType, MetaDataType]
):
    def __init__(
            self,
            protocol: BinaryTreeBalancingProtocol[ActiveKeyType, MetaDataType, 'ActiveBinaryTree'],
            reference: NullableReference[BinaryTree[KeyMetadata[ActiveKeyType, MetaDataType]]]
    ):
        assert isinstance(protocol, BinaryTreeBalancingProtocol)
        assert isinstance(reference, NullableReference)
        super().__init__(reference)
        self.protocol = protocol
        self.comparator = protocol.comparator
        assert isinstance(self.comparator, Comparator)
        self.creation = ActiveCreationProtocol(protocol)
        assert isinstance(self.creation, BinaryTreeCreationProtocol)

    @classmethod
    def empty(
            cls,
            protocol: BinaryTreeBalancingProtocol[ActiveKeyType, MetaDataType, 'ActiveBinaryTree'],
            factory: RainbowFactory[ActiveKeyType]
    ):
        assert isinstance(protocol, BinaryTreeBalancingProtocol)
        assert isinstance(factory, RainbowFactory)
        return cls(
            protocol,
            NullableReference(
                Null(),
                BinaryTreeFactory(KeyMetadataFactory(factory, protocol.empty_metadata().factory))
            )
        )

    def create(
            self,
            reference: NullableReference[BinaryTree[KeyMetadata[ActiveKeyType, MetaDataType]]]
    ) -> 'ActiveBinaryTree[ActiveKeyType, MetaDataType]':
        assert isinstance(reference, NullableReference)
        return ActiveBinaryTree(
            self.protocol,
            reference
        )

    def add(self, key: HashPoint[ActiveKeyType]) -> 'ActiveBinaryTree[ActiveKeyType, MetaDataType]':
        assert isinstance(key, HashPoint)
        if self.creation.splittable(self):
            treel: ActiveBinaryTree[ActiveKeyType, MetaDataType]
            original: HashPoint[ActiveKeyType]
            treer: ActiveBinaryTree[ActiveKeyType, MetaDataType]
            treel, original, _, treer = self.creation.split(self)
            assert isinstance(treel, ActiveBinaryTree)
            assert isinstance(original, HashPoint)
            assert isinstance(treer, ActiveBinaryTree)
            comparison: Comparison = self.comparator.compare(original, key)
            assert isinstance(comparison, Comparison)
            if isinstance(comparison, Replace):
                return self.creation.tree(treel, treer, key)
            elif isinstance(comparison, Left):
                return self.creation.tree(treel.add(key), treer, original)
            elif isinstance(comparison, Right):
                return self.creation.tree(treel, treer.add(key), original)
            else:
                raise TypeError
        else:
            return self.creation.tree(self, self, key)

    def pop(self) -> tuple['ActiveBinaryTree[ActiveKeyType, MetaDataType]', HashPoint[ActiveKeyType]]:
        assert self.creation.splittable(self)
        treel: ActiveBinaryTree[ActiveKeyType, MetaDataType]
        original: HashPoint[ActiveKeyType]
        treer: ActiveBinaryTree[ActiveKeyType, MetaDataType]
        treel, original, _, treer = self.creation.split(self)
        assert isinstance(treel, ActiveBinaryTree)
        assert isinstance(original, HashPoint)
        assert isinstance(treer, ActiveBinaryTree)
        if not self.creation.splittable(treel):
            return treer, original
        elif not self.creation.splittable(treer):
            return treel, original
        else:
            treel, key = treel.pop()
            assert isinstance(treel, ActiveBinaryTree)
            assert isinstance(key, HashPoint)
            return self.creation.tree(treel, treer, key), original

    def remove(self, key: HashPoint[ActiveKeyType]) -> 'ActiveBinaryTree[ActiveKeyType, MetaDataType]':
        assert isinstance(key, HashPoint)
        if self.creation.splittable(self):
            treel: ActiveBinaryTree[ActiveKeyType, MetaDataType]
            original: HashPoint[ActiveKeyType]
            treer: ActiveBinaryTree[ActiveKeyType, MetaDataType]
            treel, original, _, treer = self.creation.split(self)
            assert isinstance(treel, ActiveBinaryTree)
            assert isinstance(original, HashPoint)
            assert isinstance(treer, ActiveBinaryTree)
            comparison: Comparison = self.comparator.compare(original, key)
            assert isinstance(comparison, Comparison)
            if isinstance(comparison, Equal):
                return self.pop()[0]
            elif isinstance(comparison, Left):
                return self.creation.tree(treel.remove(key), treer, original)
            elif isinstance(comparison, Right):
                return self.creation.tree(treel, treer.remove(key), original)
            else:
                raise TypeError
        else:
            return self

    def query_tree(self) -> QueryCollectionInterface[ActiveKeyType]:
        return KeyMetadataQueryCollection(
            self.protocol.empty_metadata(),
            QueryBinaryTree(KeyedComparator(self.comparator), self.reference)
        )


class ActiveCreationProtocol(BalancedTreeCreationProtocol[ActiveKeyType, MetaDataType, ActiveBinaryTree]):
    def splittable(self, tree: ActiveBinaryTree[ActiveKeyType, MetaDataType]) -> bool:
        assert isinstance(tree, ActiveBinaryTree)
        return isinstance(tree.reference.reference, NotNull)

    def split(
            self,
            tree: ActiveBinaryTree[ActiveKeyType, MetaDataType]
    ) -> tuple[
        ActiveBinaryTree[ActiveKeyType, MetaDataType],
        HashPoint[ActiveKeyType],
        HashPoint[MetaDataType],
        ActiveBinaryTree[ActiveKeyType, MetaDataType]
    ]:
        assert isinstance(tree, ActiveBinaryTree)
        assert isinstance(tree.reference.reference, NotNull)
        hash_point: HashPoint[BinaryTree[KeyMetadata[ActiveKeyType, MetaDataType]]] = tree.reference.reference.value
        assert isinstance(hash_point, HashPoint)
        resolved: BinaryTree[KeyMetadata[ActiveKeyType, MetaDataType]] = hash_point.resolve()
        assert isinstance(resolved, BinaryTree)
        key_metadata: KeyMetadata[ActiveKeyType, MetaDataType] = resolved.key.resolve()
        assert isinstance(key_metadata, KeyMetadata)
        return tree.create(resolved.treel), key_metadata.key, key_metadata.metadata, tree.create(resolved.treer)

    def _tree(
            self,
            treel: ActiveBinaryTree,
            treer: ActiveBinaryTree,
            key: HashPoint[ActiveKeyType]
    ) -> ActiveBinaryTree:
        assert isinstance(treel, ActiveBinaryTree)
        assert isinstance(treer, ActiveBinaryTree)
        assert isinstance(key, HashPoint)
        return ActiveBinaryTree(
            self.protocol,
            NullableReference.off(
                BinaryTree(
                    treel.reference,
                    treer.reference,
                    HashPoint.of(KeyMetadata(key, self.protocol.metadata(treel, treer, key, self)))
                )
            )
        )
