from typing import Iterable, Optional

from rainbowadn.data.atomic.integer import Integer
from rainbowadn.hashing.hash_point_format import hash_point_format, tabulate
from rainbowadn.hashing.hashmentionable import HashMentionable
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.localmetaorigin import LocalMetaOrigin
from rainbowadn.hashing.localorigin import LocalOrigin
from rainbowadn.hashing.rainbow_factory import RainbowFactory
from rainbowadn.hashing.recursivementionable import RecursiveMentionable
from rainbowadn.hashing.resolverorigin import ResolverOrigin
from rainbowadn.wrisbt.weakreferenceindexsetbtree import WeakReferenceIndexSetBTree, WrisbtFactory
from rainbowadn.wrisbt.wrisbtparametres import WrisbtParametres

__all__ = ('WrisbtRoot', 'WrisbtRootFactory',)


class WrisbtRoot(RecursiveMentionable):
    def __init__(self, root: HashPoint[WeakReferenceIndexSetBTree], height: int, parametres: WrisbtParametres):
        assert isinstance(root, HashPoint)
        assert isinstance(height, int)
        assert isinstance(parametres, WrisbtParametres)
        self.root = root
        self.height = height
        self.parametres = parametres

    def points(self) -> Iterable[HashPoint]:
        return [self.root]

    def __bytes__(self):
        return bytes(self.root) + bytes(Integer(self.height))

    def __factory__(self) -> RainbowFactory['WrisbtRoot']:
        return WrisbtRootFactory(self.parametres)

    @classmethod
    def empty(cls, parametres: WrisbtParametres) -> 'WrisbtRoot':
        assert isinstance(parametres, WrisbtParametres)
        return WrisbtRoot(HashPoint.of(WeakReferenceIndexSetBTree(b'', 0, parametres, True, ())), 0, parametres)

    def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return f'(root)' \
               f'{tabulate(tab)}{self.height}' \
               f'{tabulate(tab)}{hash_point_format(self.root, tab)}'

    def contains(self, key: bytes) -> bool:
        assert isinstance(key, bytes)
        assert len(key) == self.parametres.keysize

        root: WeakReferenceIndexSetBTree = self.root.resolve()
        assert isinstance(root, WeakReferenceIndexSetBTree)

        return root.contains(key)

    def add(self, key: bytes) -> 'WrisbtRoot':
        assert isinstance(key, bytes)
        assert len(key) == self.parametres.keysize

        root: WeakReferenceIndexSetBTree = self.root.resolve().add(key)
        assert isinstance(root, WeakReferenceIndexSetBTree)

        if root.full():
            left, middle, right = root.split()
            assert isinstance(left, WeakReferenceIndexSetBTree)
            assert isinstance(middle, bytes)
            assert isinstance(right, WeakReferenceIndexSetBTree)
            root = WeakReferenceIndexSetBTree(
                middle + bytes(HashPoint.of(left)) + bytes(HashPoint.of(right)),
                root.height + 1,
                root.parametres,
                True,
                (LocalMetaOrigin(LocalOrigin(left)), LocalMetaOrigin(LocalOrigin(right)))
            )
            assert isinstance(root, WeakReferenceIndexSetBTree)
        return self.of(root)

    @classmethod
    def of(cls, root: WeakReferenceIndexSetBTree) -> 'WrisbtRoot':
        assert isinstance(root, WeakReferenceIndexSetBTree)
        return cls(HashPoint.of(root), root.height, root.parametres)

    def keys(self) -> list[bytes]:
        return list(self.root.resolve().iter_keys())

    def index(
            self, target: HashPoint, exclude: Optional['WrisbtRoot']
    ) -> 'WrisbtRoot':
        assert isinstance(target, HashPoint)
        if exclude is None:
            exclude = self.empty(self.parametres)
        assert isinstance(exclude, WrisbtRoot)
        key: bytes = target.point
        assert isinstance(key, bytes)
        if exclude.contains(key) or self.contains(key):
            return self
        tree = self
        value: HashMentionable = target.resolve()
        assert isinstance(value, HashMentionable)
        if isinstance(value, RecursiveMentionable):
            for hash_point in value.points():
                tree = tree.index(hash_point, exclude)
        tree = tree.add(key)
        return tree


class WrisbtRootFactory(RainbowFactory[WrisbtRoot]):
    def __init__(self, parametres: WrisbtParametres):
        assert isinstance(parametres, WrisbtParametres)
        self.parametres = parametres

    def from_bytes(self, source: bytes, resolver: HashResolver) -> WrisbtRoot:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        height: int = Integer.from_bytes(source[HashPoint.HASH_LENGTH:], resolver).integer
        assert isinstance(height, int)
        return WrisbtRoot(
            ResolverOrigin(
                WrisbtFactory(height, self.parametres, True), source[:HashPoint.HASH_LENGTH], resolver
            ).hash_point(),
            height,
            self.parametres
        )
